#pragma comment(linker, "/version:2.0") // MUST be in the form of major.minor

#include <windows.h>
#include <shlwapi.h>
#include <stdio.h>
#include <tchar.h>
#include "libs\SimpleString.h"

#define STR_TC_EXT TEXT(".timeclone")
#define CCH_TC_EXT 10        // SSLen(STR_TC_EXT) == CCH_TC_EXT
#define CB_SCRATCH 0x1000    // One page ought to be big enough
#define CB_TSTAMPS 51        // 3 * (64 bits * (1 byte / 4 bits) + 1 byte)

static const BYTE UTF8_BOM[] = { 0xEF, 0xBB, 0xBF };

#ifdef UNICODE
#define TStrToUTF8(t, a, c) WideCharToMultiByte(CP_UTF8, 0, t, -1, a, c, NULL, NULL)
#define UTF8ToTStr(a, t, c) MultiByteToWideChar(CP_UTF8, 0, a, -1, t, c)
#else
#error Unicode is required.
#endif

typedef struct {
	BOOL fVerbose;           // print status?
	enum {
		MODE_NORMAL = 0,
		MODE_EXPORT,
		MODE_IMPORT
	} mode;
	UINT_PTR cchPathPrefix;  // [IMPORT/EXPORT] length of the path prefix
	PVOID pvScratch;         // [IMPORT/EXPORT] scratch buffer
	PBYTE pbFileData;        // [IMPORT] file data buffer
	PBYTE pbFileDataEnd;     // [IMPORT] end of the file data buffer
	HANDLE hFile;            // [EXPORT] handle of the .timeclone file
	UINT cSuccesses;         // successfully cloned timestamps
	UINT cErrors;            // errors encountered
} WORKCTX, *PWORKCTX;

typedef union {
	struct {
		union {
			FILETIME ftCreationTime;
			UINT64 uiCreationTime;
		};
		union {
			FILETIME ftLastAccessTime;
			UINT64 uiLastAccessTime;
		};
		union {
			FILETIME ftLastWriteTime;
			UINT64 uiLastWriteTime;
		};
	};
	UINT64 uiTime[3];
} FILETIMEEX, *PFILETIMEEX;

VOID WINAPI CloneTimeItem( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILETIMEEX pftex );
VOID WINAPI CloneTimeTree( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath );
__forceinline BOOL InitTimeCloneFile( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath );
__forceinline VOID ParseTimeCloneFile( PWORKCTX pctx, PCTSTR pszTargetPath );
__forceinline BOOL WriteData( HANDLE hFile, LPCVOID pvBuffer, DWORD cbBuffer );
VOID __fastcall NormalizeSlashes( PTSTR pszPath );
VOID __fastcall PrintError( PCTSTR pszPath, DWORD dwErrorCode );
__forceinline BOOL IsFlag( PCTSTR pszArg );
__forceinline BOOL CheckFlagI( PCTSTR pszArg, TCHAR ch );
__forceinline BOOL IsSpecialDirectoryName( PCTSTR pszPath );
__forceinline BOOL IsValidDirectory( PCTSTR pszPath );
__forceinline BOOL IsValidFile( PCTSTR pszPath );

UINT __fastcall EntryMain( UINT argc, PTSTR *argv )
{
	PTSTR pszSourcePath = NULL;
	PTSTR pszTargetPath = NULL;
	BOOL fRecurse = FALSE;
	BOOL fShowUsage = FALSE;

	// Parse command line parameters

	BOOL fFlagsDone = FALSE;
	UINT i;

	WORKCTX ctx;
	ZeroMemory(&ctx, sizeof(ctx));

	for (i = 1; i < argc; ++i)
	{
		if (fFlagsDone || !IsFlag(argv[i]))
		{
			fFlagsDone = TRUE;

			if (pszSourcePath == NULL)
				pszSourcePath = argv[i];
			else if (pszTargetPath == NULL)
				pszTargetPath = argv[i];
			else
				fShowUsage = TRUE;
		}
		else if (!ctx.fVerbose && CheckFlagI(argv[i], TEXT('v')))
			ctx.fVerbose = TRUE;
		else if (ctx.mode == MODE_NORMAL && CheckFlagI(argv[i], TEXT('e')))
			ctx.mode = MODE_EXPORT;
		else if (ctx.mode == MODE_NORMAL && CheckFlagI(argv[i], TEXT('i')))
			ctx.mode = MODE_IMPORT;
		else if (!fRecurse && CheckFlagI(argv[i], TEXT('r')))
			fRecurse = TRUE;
		else
			fShowUsage = TRUE;
	}

	// Act according to the command line parameters

	if (fShowUsage || pszTargetPath == NULL)
	{
		static const TCHAR szUsageTemplateGeneral[] = TEXT("  -%c  %s.\n");

		_tprintf(TEXT("Usage: timeclone [-v] [-e | -i] [-r] source target\n\n"));
		_tprintf(TEXT("Options:\n"));
		_tprintf(szUsageTemplateGeneral, TEXT('v'), TEXT("Verbose mode; prints status"));
		_tprintf(szUsageTemplateGeneral, TEXT('e'), TEXT("Export mode; target is a .timeclone file"));
		_tprintf(szUsageTemplateGeneral, TEXT('i'), TEXT("Import mode; source must be a .timeclone file"));
		_tprintf(szUsageTemplateGeneral, TEXT('r'), TEXT("Recursive mode; processes directory trees"));

		return(~1);
	}
	else
	{
		NormalizeSlashes(pszSourcePath);
		NormalizeSlashes(pszTargetPath);

		// Open and load the .timeclone file for import/export
		if (ctx.mode != MODE_NORMAL && !InitTimeCloneFile(&ctx, pszSourcePath, pszTargetPath))
			return(1);

		if (ctx.mode == MODE_IMPORT)
			ParseTimeCloneFile(&ctx, pszTargetPath);
		else
		{
			CloneTimeItem(&ctx, pszSourcePath, pszTargetPath, NULL);

			if (fRecurse && IsValidDirectory(pszSourcePath) && (ctx.mode == MODE_EXPORT || IsValidDirectory(pszTargetPath)))
				CloneTimeTree(&ctx, pszSourcePath, pszTargetPath);
		}

		// Cleanup
		if (ctx.pvScratch) LocalFree(ctx.pvScratch);
		if (ctx.pbFileData) LocalFree(ctx.pbFileData);
		if (ctx.hFile) CloseHandle(ctx.hFile);

		// Post-mortem
		_tprintf(TEXT("\t%d timestamp(s) cloned"), ctx.cSuccesses);
		if (ctx.cErrors) _tprintf(TEXT(", with %d error(s)"), ctx.cErrors);
		_tprintf(TEXT(".\n"));
	}

	return(ctx.cErrors);
}

VOID WINAPI CloneTimeItem( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath, PFILETIMEEX pftex )
{
	BOOL fSuccess = FALSE;
	FILETIMEEX ftex;

	if (pctx->fVerbose)
		_tprintf(TEXT("%s\n"), pszSourcePath);

	// Load the source timestamp, if it was not provided

	if (pftex == NULL)
	{
		HANDLE hFile = CreateFile(
			pszSourcePath,
			0,
			FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
			NULL,
			OPEN_EXISTING,
			FILE_FLAG_BACKUP_SEMANTICS,
			NULL
		);

		if (hFile != INVALID_HANDLE_VALUE)
		{
			if (GetFileTime(hFile, &ftex.ftCreationTime, &ftex.ftLastAccessTime, &ftex.ftLastWriteTime))
				pftex = &ftex;

			CloseHandle(hFile);
		}
	}

	if (pftex)
	{
		if (pctx->mode != MODE_EXPORT)
		{
			// Normal or import modes: Set the target file timestamp

			HANDLE hFile = CreateFile(
				pszTargetPath,
				GENERIC_WRITE,
				FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,
				NULL,
				OPEN_EXISTING,
				FILE_FLAG_BACKUP_SEMANTICS,
				NULL
			);

			if (hFile != INVALID_HANDLE_VALUE)
			{
				fSuccess = SetFileTime(
					hFile,
					&pftex->ftCreationTime,
					&pftex->ftLastAccessTime,
					&pftex->ftLastWriteTime
				);

				CloseHandle(hFile);
			}
		}
		else
		{
			// Export mode: Save the timestamp to the open file handle

			PBYTE pbBuffer = pctx->pvScratch;

			// Format timestamps

			sprintf(
				pbBuffer,
				"%016I64X|%016I64X|%016I64X|",
				pftex->uiCreationTime,
				pftex->uiLastAccessTime,
				pftex->uiLastWriteTime
			);

			pbBuffer += CB_TSTAMPS;

			// Convert the filename to UTF-8

			pbBuffer += TStrToUTF8(pszSourcePath + pctx->cchPathPrefix, pbBuffer, CB_SCRATCH - CB_TSTAMPS - sizeof(WORD)) - 1;

			// Add a newline

			*((UPWORD)pbBuffer) = 0x0A0D;
			pbBuffer += sizeof(WORD);

			// Write the line

			if (WriteData(pctx->hFile, pctx->pvScratch, (DWORD)(pbBuffer - (PBYTE)pctx->pvScratch)))
				fSuccess = TRUE;
		}
	}

	if (fSuccess)
	{
		++pctx->cSuccesses;
	}
	else
	{
		++pctx->cErrors;
		PrintError(pszSourcePath, GetLastError());
	}
}

VOID WINAPI CloneTimeTree( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath )
{
	HANDLE hFind;
	WIN32_FIND_DATA finddata;

	// Copy the paths to a scratchpad buffer
	TCHAR szSourcePath[MAX_PATH << 2];
	TCHAR szTargetPath[MAX_PATH << 2];
	PTSTR pszSourcePathAppend = SSChainNCpy(szSourcePath, pszSourcePath, SSLen(pszSourcePath));
	PTSTR pszTargetPathAppend = SSChainNCpy(szTargetPath, pszTargetPath, SSLen(pszTargetPath));

	// Ensure that we have a slash termination
	*pszSourcePathAppend++ = TEXT('\\');
	*pszTargetPathAppend++ = TEXT('\\');

	SSCpy2Ch(pszSourcePathAppend, TEXT('*'), 0);

	if ((hFind = FindFirstFile(szSourcePath, &finddata)) != INVALID_HANDLE_VALUE)
	{
		do
		{
			if (!IsSpecialDirectoryName(finddata.cFileName))
			{
				SSCpy(pszSourcePathAppend, finddata.cFileName);
				SSCpy(pszTargetPathAppend, finddata.cFileName);

				CloneTimeItem(pctx, szSourcePath, szTargetPath, (PFILETIMEEX)&finddata.ftCreationTime);

				if (finddata.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
					CloneTimeTree(pctx, szSourcePath, szTargetPath);
			}

		} while (FindNextFile(hFind, &finddata));

		FindClose(hFind);
	}
	else
	{
		++pctx->cErrors;
		PrintError(pszSourcePath, GetLastError());
	}
}

__forceinline BOOL InitTimeCloneFile( PWORKCTX pctx, PCTSTR pszSourcePath, PCTSTR pszTargetPath )
{
	TCHAR szPath[MAX_PATH << 2];
	HANDLE hFile;

	PCTSTR pszPath = (pctx->mode == MODE_EXPORT) ? pszTargetPath : pszSourcePath;
	pctx->cchPathPrefix = SSLen((pctx->mode == MODE_EXPORT) ? pszSourcePath : pszTargetPath);

	// When exporting, ensure that the .timeclone extension is always there, and
	// when importing, tack on a missing .timeclone extension if and only if the
	// input file can be found with the extension and not without.

	if (pctx->mode == MODE_EXPORT || !IsValidFile(pszPath))
	{
		UINT_PTR cchPath = SSLen(pszPath);

		if (cchPath < CCH_TC_EXT || StrCmpI(pszPath + cchPath - CCH_TC_EXT, STR_TC_EXT))
		{
			SSStaticCpy(SSChainNCpy(szPath, pszPath, cchPath), STR_TC_EXT);

			if (pctx->mode == MODE_EXPORT || IsValidFile(szPath))
				pszPath = szPath;
		}
	}

	// Open the file

	hFile = CreateFile(
		pszPath,
		(pctx->mode == MODE_EXPORT) ? GENERIC_WRITE : GENERIC_READ,
		FILE_SHARE_READ,
		NULL,
		(pctx->mode == MODE_EXPORT) ? CREATE_ALWAYS : OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
		NULL
	);

	if (hFile != INVALID_HANDLE_VALUE)
	{
		// Allocate a scratch buffer

		if (pctx->pvScratch = LocalAlloc(LMEM_FIXED, CB_SCRATCH))
		{
			if (pctx->mode == MODE_EXPORT)
			{
				// Note that pctx->hFile will always be either NULL (its initial
				// state) or a valid handle; it is never INVALID_HANDLE_VALUE

				if (WriteData(hFile, UTF8_BOM, sizeof(UTF8_BOM)))
				{
					pctx->hFile = hFile;
					return(TRUE);
				}
			}
			else
			{
				// Load the file for import; we will process it later

				LARGE_INTEGER cbFileData;
				DWORD cbBytesRead;

				if ( (GetFileSizeEx(hFile, &cbFileData)) &&
				     (pctx->pbFileData = LocalAlloc(LMEM_FIXED, cbFileData.LowPart + sizeof(DWORD))) &&
				     (ReadFile(hFile, pctx->pbFileData, cbFileData.LowPart, &cbBytesRead, NULL)) &&
				     (cbFileData.LowPart == cbBytesRead) )
				{
					pctx->pbFileDataEnd = pctx->pbFileData + cbFileData.LowPart;
					*((UPDWORD)pctx->pbFileDataEnd) = 0;
					CloseHandle(hFile);
					return(TRUE);
				}

				// Once we fall to this point, we have failed, so we should
				// start cleaning up

				if (pctx->pbFileData)
					LocalFree(pctx->pbFileData);
			}

			LocalFree(pctx->pvScratch);
		}

		CloseHandle(hFile);
	}

	PrintError(pszPath, GetLastError());
	return(FALSE);
}

__forceinline VOID ParseTimeCloneFile( PWORKCTX pctx, PCTSTR pszTargetPath )
{
	FILETIMEEX ftex;
	UINT i, j;
	BOOL fValid;

	PSTR pszFileNameUTF8;
	PTSTR pszFileName = pctx->pvScratch;
	PTSTR pszFileNameAppend = SSChainNCpy(pszFileName, pszTargetPath, pctx->cchPathPrefix);
	PBYTE pbData = pctx->pbFileData;

	// Note that the allocated buffer is at least sizeof(DWORD) larger than
	// the actual length of the file data, which allows for bounds checking to
	// be skipped in certain instances.

	if (!memcmp(pbData, UTF8_BOM, sizeof(UTF8_BOM)))
		pbData += sizeof(UTF8_BOM);

	while (pbData < pctx->pbFileDataEnd)
	{
		// Read and validate the timestamps

		ZeroMemory(&ftex, sizeof(ftex));
		fValid = TRUE;

		for (i = 0; i < 3 && pbData < pctx->pbFileDataEnd; ++i)
		{
			for (j = 0; j < 16 && pbData < pctx->pbFileDataEnd; ++j)
			{
				ftex.uiTime[i] <<= 4;
				*pbData |= 0x20;

				if (*pbData >= '0' && *pbData <= '9')
					ftex.uiTime[i] += *pbData - '0';
				else if (*pbData >= 'a' && *pbData <= 'f')
					ftex.uiTime[i] += *pbData - ('a' - 10);
				else
					fValid = FALSE;

				++pbData;
			}

			if (*pbData++ != '|')
				fValid = FALSE;
		}

		// Determine the extent of the filename

		pszFileNameUTF8 = pbData;

		while (pbData < pctx->pbFileDataEnd && *pbData != 0x0A && *pbData != 0x0D)
			++pbData;

		*pbData++ = 0;

		// Get filename and set timestamp

		if (fValid)
		{
			UTF8ToTStr(
				pszFileNameUTF8,
				pszFileNameAppend,
				CB_SCRATCH / sizeof(TCHAR) - (UINT)pctx->cchPathPrefix
			);

			CloneTimeItem(pctx, pszFileName, pszFileName, &ftex);
		}

		// Go to the next line

		while (pbData < pctx->pbFileDataEnd && (*pbData == 0x0A || *pbData == 0x0D))
			++pbData;
	}
}

__forceinline BOOL WriteData( HANDLE hFile, LPCVOID pvBuffer, DWORD cbBuffer )
{
	DWORD cbWritten;
	return(WriteFile(hFile, pvBuffer, cbBuffer, &cbWritten, NULL) && cbBuffer == cbWritten);
}

VOID __fastcall NormalizeSlashes( PTSTR pszPath )
{
	PTSTR psz = pszPath;

	while (*psz)
	{
		if (*psz == TEXT('/'))
			*psz = TEXT('\\');

		++psz;
	}

	while (--psz >= pszPath && *psz == TEXT('\\'))
		*psz = 0;
}

VOID __fastcall PrintError( PCTSTR pszPath, DWORD dwErrorCode )
{
	_ftprintf(stderr, TEXT("Error: %s"), pszPath);

	if (dwErrorCode)
	{
		PTSTR pszErrorMessage = NULL;

		DWORD cchMessage = FormatMessage(
			FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
			NULL,
			dwErrorCode,
			0,
			(PTSTR)&pszErrorMessage,
			0,
			NULL
		);

		if (cchMessage)
		{
			pszErrorMessage[StrCSpn(pszErrorMessage, TEXT("\r\n"))] = 0;
			_ftprintf(stderr, TEXT(" [%s]"), pszErrorMessage);
		}
		else
		{
			_ftprintf(stderr, TEXT(" [%08X]"), dwErrorCode);
		}

		LocalFree(pszErrorMessage);
	}

	_ftprintf(stderr, TEXT("\n"));
}

__forceinline BOOL IsFlag( PCTSTR pszArg )
{
	return(
		(pszArg[0] | 0x02) == TEXT('/') &&
		(pszArg[1]       ) != 0 &&
		(pszArg[2]       ) == 0
	);
}

__forceinline BOOL CheckFlagI( PCTSTR pszArg, TCHAR ch )
{
	return(
		(pszArg[1] | 0x20) == ch
	);
}

__forceinline BOOL IsSpecialDirectoryName( PCTSTR pszPath )
{
	// TRUE if name is "." or ".."

	#ifdef UNICODE
	return(
		(*((UPDWORD)pszPath) == WCHARS2DWORD(L'.', 0)) ||
		(*((UPDWORD)pszPath) == WCHARS2DWORD(L'.', L'.') && pszPath[2] == 0)
	);
	#else
	return(
		(*((UPWORD)pszPath) == CHARS2WORD('.', 0)) ||
		(*((UPWORD)pszPath) == CHARS2WORD('.', '.') && pszPath[2] == 0)
	);
	#endif
}

__forceinline BOOL IsValidDirectory( PCTSTR pszPath )
{
	DWORD dwAttributes = GetFileAttributes(pszPath);
	return(dwAttributes != INVALID_FILE_ATTRIBUTES && dwAttributes & FILE_ATTRIBUTE_DIRECTORY);
}

__forceinline BOOL IsValidFile( PCTSTR pszPath )
{
	// Since INVALID_HANDLE_VALUE is -1, if the FILE_ATTRIBUTE_DIRECTORY bit is
	// set, then the item is either nonexistent or is a directory--not a file.
	return(!(GetFileAttributes(pszPath) & FILE_ATTRIBUTE_DIRECTORY));
}
